package ru.dbotthepony.mc.otm.core

import com.mojang.math.Quaternion
import com.mojang.math.Vector3f
import net.minecraft.world.phys.Vec3
import kotlin.math.acos
import kotlin.math.cos
import kotlin.math.sin

typealias Vector = Vec3

val VECTOR_UP = Vector(0.0, 1.0, 0.0)
val VECTOR_FORWARD = Vector(1.0, 0.0, 0.0)
val VECTOR_BACKWARD = Vector(1.0, 0.0, 0.0)
val VECTOR_DOWN = Vector(0.0, -1.0, 0.0)
val VECTOR_RIGHT = Vector(0.0, 0.0, 1.0)
val VECTOR_LEFT = Vector(0.0, 0.0, -1.0)

fun Vector.asAngle(): Angle {
	val norm = normalize()

	if (norm.x < 0) {
		return Angle(acos(norm.dot(VECTOR_UP)) - Angle.PI_HALF, -acos(Vector(norm.x, 0.0, norm.z).normalize().dot(VECTOR_RIGHT)))
	}

	return Angle(acos(norm.dot(VECTOR_UP)) - Angle.PI_HALF, acos(Vector(norm.x, 0.0, norm.z).normalize().dot(VECTOR_RIGHT)))
}

operator fun Vector.plus(direction: Vector): Vector = this.add(direction)
operator fun Vector.minus(direction: Vector): Vector = this.subtract(direction)
operator fun Vector.unaryMinus(): Vector = Vector(-x, -y, -z)
operator fun Vector.component1() = x
operator fun Vector.component2() = y
operator fun Vector.component3() = z
operator fun Vector.times(v: Double): Vector = this.multiply(v, v, v)
operator fun Vector.div(v: Double): Vector {
	val inv = 1.0 / v
	return this.multiply(inv, inv, inv)
}

fun Vector.left() = asAngle().left() * this.length()
fun Vector.right() = asAngle().right() * this.length()
fun Vector.up() = asAngle().up() * this.length()
fun Vector.down() = asAngle().down() * this.length()

fun Vector.rotateAroundAxis(axis: Vector, rotation: Double): Vector {
	return this * cos(rotation) + axis.cross(this) * sin(rotation) + axis * axis.dot(this) * (1 - cos(rotation))
}

fun Vector.rotate(angle: Angle): Vector {
	val rotatedYaw = this.rotateAroundAxis(VECTOR_UP, angle.yaw)
	return rotatedYaw.rotateAroundAxis(rotatedYaw.asAngle().left(), angle.pitch)
}

fun Vector.asVector3f(): Vector3f {
	return Vector3f(x.toFloat(), y.toFloat(), z.toFloat())
}

fun Quaternion(vec: Vector, rotation: Float, isDegrees: Boolean): Quaternion {
	return Quaternion(vec.asVector3f(), rotation, isDegrees)
}

fun Vector(x: Double): Vector {
	return Vector(x, 0.0, 0.0)
}

fun Vector(x: Double, y: Double): Vector {
	return Vector(x, y, 0.0)
}

fun Vector.rotate(q: Quaternion): Vector {
	val quaternion = Quaternion(q)
	quaternion.mul(Quaternion(x.toFloat(), y.toFloat(), z.toFloat(), 0.0f))
	val quaternion1 = Quaternion(q)
	quaternion1.conj()
	quaternion.mul(quaternion1)
	return Vector(quaternion.i().toDouble(), quaternion.j().toDouble(), quaternion.k().toDouble())
}

data class Angle(val pitch: Double = 0.0, val yaw: Double = 0.0, val roll: Double = 0.0) {
	fun forward(): Vector {
		val yaw = yaw
		val x = cos(yaw) * cos(pitch)
		val y = -sin(pitch)
		val z = sin(yaw) * cos(pitch)

		return Vector(x, y, z)
	}

	fun left(): Vector {
		return -right()
	}

	fun right(): Vector {
		val yaw = yaw + PI_HALF
		val pitch = 0.0
		val x = cos(yaw) * cos(pitch)
		val y = -sin(pitch)
		val z = sin(yaw) * cos(pitch)

		return Vector(x, y, z)
	}

	fun up(): Vector {
		val yaw = yaw - PI_HALF
		val pitch = pitch - PI_HALF
		val x = cos(yaw) * cos(pitch)
		val y = -sin(pitch)
		val z = sin(yaw) * cos(pitch)

		return Vector(x, y, z)
	}

	fun down(): Vector {
		val yaw = yaw - PI_HALF
		val pitch = pitch + PI_HALF
		val x = cos(yaw) * cos(pitch)
		val y = -sin(pitch)
		val z = sin(yaw) * cos(pitch)

		return Vector(x, y, z)
	}

	companion object {
		const val PI_HALF = Math.PI / 2.0

		fun degrees(pitch: Double, yaw: Double, roll: Double): Angle {
			return Angle(Math.toDegrees(pitch), Math.toDegrees(yaw), Math.toDegrees(roll))
		}
	}
}
